<?php
/*  Copyright 2008  Michael J. Walker  (email : mike@englishmike.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

add_action('publish_page', 'az_cache_post_changed');
add_action('publish_post', 'az_cache_post_changed');
add_action('save_post', 'az_cache_post_changed');
add_action('deleted_post', 'az_cache_post_changed');
add_action('wp_update_post', 'az_cache_post_changed');
add_action('wp_insert_post', 'az_cache_post_changed');
add_action('check_ajax_referer', 'az_cache_check_custom');

//add_action('wp_ajax_addmeta', 'az_cache_check');
//add_action('edit_post', 'az_cache_check');
//add_action('init', 'az_cache_check');

/**
 * Creates a new cached object for the specified index.
 *
 * @param az_request $index the current index
 * @param int $pageid current page id displaying index (can be a post)
 * @param int $pageno current page number of index being displayed
 * @return az_index_cache cached information for current index
 */
function az_cache_get($index, $pageid, $pageno) {
	return new az_index_cache($index, $pageid, $pageno);
}

/**
 * Main entry point for processing changed posts and pages.  Hooks into 
 * functions that modify posts and pages in some way call here so that
 * the posts can be added to the dirty list of all the indexes.  Note
 * that we have to filter out changes to revisions by checking the post_type/
 *
 * @param int $postid id of post changed 
 */
function az_cache_post_changed($postid) {
    $post = get_post($postid);
    if ($post->post_type == 'post' || $post->post_type == 'page') {
        //az_trace("-----------------------------------------------");
        //az_trace("az_cache_post_changed:".current_filter()." - postid = $postid");
        //az_trace("az_cache_post_changed: - post_type = $post->post_type");
        az_add_dirty_post($postid);
    }
}

/**
 * There is no easy way to capture changes to custom fields that are done via AJAX calls.
 * This callback is called *before* any changes are made, so while we can log that a
 * change has been made (which we do) there is no way to find out what the change is
 * and whether it affects the index or not.  There is no consistency in the way the 
 * postid is passed on the AJAX call, which is the reason for the convoluted code.
 *
 * @param object $action - not used
 */
function az_cache_check_custom($action) {
	global $wpdb;
//	az_trace("az_cache_check:".current_filter()." - action = $action");
    $action = $_REQUEST['action'];
    if ($action == 'add-meta') {
        $postid = $_REQUEST['post_id'];
        if (empty($postid)) {
        	$meta = $_REQUEST['meta'];
            $idmeta = current(array_keys($meta));
        }
    } else if ($action == 'delete-meta') {
        $idmeta = $_REQUEST['id'];
    }
    
    if (!empty($idmeta)) {
        $postid = $wpdb->get_var("SELECT post_id FROM $wpdb->postmeta WHERE meta_id = '$idmeta'");
    }
    
    if (!empty($postid)) {
        $post = get_post($postid);
        if ($post->post_type == 'post' || $post->post_type == 'page') {
            //az_trace("-----------------------------------------------");
            //az_trace("az_cache_check_custom: custom field modified - postid: $postid");
            //az_trace("az_cache_check_custom: - post_type = $post->post_type");
            az_add_dirty_post($postid);
        }
    }
}

/**
 * Fetch the list of dirty posts -- and only unserialize them if
 * they haven't been unserialized already.
 *
 * @return array of dirty posts.
 */
function az_get_dirty_posts() {
    //az_trace('az_get_dirty_posts');
    $dirtyposts = get_option("az_cache_dirty");
    if (is_serialized($dirtyposts)) {
        $dirtyposts = unserialize($dirtyposts); 
    }
    return $dirtyposts;
}

function az_set_dirty_posts($dirtyposts) {
	update_option("az_cache_dirty", serialize($dirtyposts));
	//az_trace("az_set_dirty_posts: saving dirtyposts:");
}

function az_add_dirty_post($postid) {
	az_acquire_cache_mutex();
    $dirtyposts = az_get_dirty_posts();
    // Only process the dirty post if its the same post as before if
    // a quarter of a second or more has elapse, otherwise we will
    // process the same changes multiple time. 
    $elapsed = ($postid == $dirtyposts[0][0]) ? microtime(true) - $dirtyposts[0][1] : 1;
    if ($elapsed > 0.25) {
        //az_trace("******************************************************");
        $indexes = az_get_indexes("*");
        // Add the dirty post to all the indexes since we have no idea which
        // posts are displayed in which indexes.
        foreach ($indexes as $index) {
            // Add the post to the dirty list if necessary.
            if (!(is_array($dirtyposts[$index->idindex]) && in_array(intval($postid), $dirtyposts[$index->idindex]))
 	                                                     && !az_is_set($index->options, "disable-cache")) {
 	            //az_trace("az_add_dirty_post: adding postid $postid to the dirty list of $index->idindex");
                $dirtyposts[$index->idindex][] = intval($postid);
 	        } 
            // The blog may be using an HTML cache (like WP-Super-Cache), so we 
            // have to check immediately if we need to invalidate the HTML cache
            // since we won't be able to when the cached HTML of the index itself is accessed.
            $dirtyposts = az_process_dirty_post($index, $postid, $dirtyposts);
            //az_trace("dirty: id = ".$index->idindex.": ".(implode(",", $dirtyposts[$index->idindex])));
	    }
        //az_dump_dirty($dirtyposts);
    }
	// Save the postid and timestamp for checking next time we hook a function.
    $dirtyposts[0][0] = $postid;
    $dirtyposts[0][1] = microtime(true);
    az_set_dirty_posts($dirtyposts);
    // Release the semaphore now that we are done.
    az_release_cache_mutex();
}

/**
 * For the specified index, find out if the post invalidates the 
 * cache.  If it does, then call the HTML cache function to invalidate
 * the HTML cache for the page(s) the index is being displayed on. 
 *
 * @param db_index $dbindex index settings loaded from the database
 * @param int $postid id of the post that has just been modified in some way
 */
function az_process_dirty_post($index, $postid, $dirtyposts) {
    az_trace("az_process_dirty_post: idindex = $index->idindex - postid = $postid");
    $req = new az_request();
    $req->set_vars_from_index($index);
    $cache = new az_index_cache($req, 0, 0, false);

    // First check to see if post is in index.  If it is, we assume that the
    // change might affect the index page even if it might not. 
    //az_trace("az_process_dirty_post : postid = $postid");
    //az_trace("az_process_dirty_post : cache = ".implode(",", $cache->itemcache));
    $dirty = az_is_set($index->options, 'disable-cache') || (is_array($cache->itemcache) && in_array($postid, $cache->itemcache));
    //az_trace("az_process_dirty_post: is post in index? ".($dirty ? "yes" : "no"));
    if (!$dirty) {
        // The post was not in the index, but could be a new one
        // to be added to the index.  In this case we need to run
        // the usual query against the postid to see if it should be
        // added to the index.
        $posts = $cache->query_index_items($req, array($postid), true);
        $dirty = is_array($posts) && count($posts) > 0;
        az_trace("az_process_dirty_post: is post to be added to index? ".($dirty ? "yes" : "no"));
    }
    
    // If it looks as though the index is dirty, then tell the HTML
    // cache to remove the index page.
    if ($dirty) {
        $dirtyposts = az_check_html_cache($index->idindex, $dirtyposts);
    }
    return $dirtyposts;
}

/**
 * This function must be called if a process outside the dirty mutex semaphore
 * wants to flush the cache -- for example, two calls are made from the admin
 * pages, since they do not have the semaphore lock.
 *
 * @param int $idindex id of index whose pages are to be flushed
 */
function az_flush_html_cache($idindex) {
    az_acquire_cache_mutex();
    $dirtyposts = az_get_dirty_posts();
    az_check_html_cache($idindex, $dirtyposts);
    unset($dirtyposts[$idindex]);
    az_set_dirty_posts($dirtyposts);
    az_release_cache_mutex();
} 

function az_check_html_cache($idindex, $dirtyposts) {
    az_trace("az_check_html_cache: $idindex");
    $dirtypages = $dirtyposts['pages'][$idindex];
    if (is_array($dirtypages)) {
        for ($i = 0; $i < count($dirtypages); $i++) {
            az_trace("az_check_html_cache: removing page id ".$dirtypages[$i]." from HTML cache.");
            if (function_exists('wp_cache_post_change')) {
                wp_cache_post_change($dirtypages[$i]);
            }
        }
        unset($dirtyposts['pages'][$idindex]);
    }
    return $dirtyposts;
}

function az_reset_dirty_posts($dirtyposts, $idindex) {
	$dirtyposts[$idindex] = array();
    az_set_dirty_posts($dirtyposts);
}

class az_index_cache {
	var $index;          // Current index
	var $pageno;         // Current page number
    var $pagecount;      // Number of pages in the index
	var $items;          // Array of items;
    var $itemcache;      // Cache of item ids stored in database
	var $alphalinks;     // Array of alphabetical links
	var $ignorechars;    // Characters to ignore in a sort
	var $is_multipage;   // True if a multipage index
    var $has_alphalinks; // True if alphalinks being used
    var $cache_disabled; // True if caching is disabled for this index
    
    /**
     * Constructor for the az_index_cache class.  This function either rebuilds the
     * cached index from cache data stored in the database or, if there is no stored
     * cache, or changes to the blog make the cache out of date, then the index cache
     * is built from scratch.
     *
     * @param az_request $index index to be displayed, include all the current settings
     * @param int $pageid id of page or post displaying the index
     * @param int $pageno the index of the page in the index to be displayed
     * @param boolean $buildcache if true (default) then fully build the cache 
     * @return az_index_cache the index cache
     */
    function az_index_cache($index, $pageid, $pageno, $buildcache = true) {
		$time_start = microtime(true);
		$this->cache_disabled = az_is_set($index->options, 'disable-cache');
        $this->has_alphalinks = az_is_set($index->options, 'alpha-links');
        $this->is_multipage = az_is_set($index->options, 'multipage') && $index->perpage > 0;
        $this->ignorechars = az_get_ignorechars($index);
        
        $this->index = $index;
		$this->pageno = $pageno;
        $this->itemcache = unserialize($index->itemcache);

        if ($buildcache) {
            // This is the main path for building or rebuilding the cache
            $this->add_current_page_to_cache($index->id, $pageid);
        
            if ($this->cache_disabled || $this->is_dirty()) {
       		    $this->build_index();
                $this->write_cache();
            } else {
                $this->build_index_from_cache();
            }
        }
        $time_end = microtime(true);
        //az_println("Cache execution in: "+(($time_end - $time_start)*1000)."ms");
	}

	/**
	 * Build the index from scratch.  All items in the index need to be loaded from the database
	 * and then sorted before the correct page of the index and its contents can be display. This
	 * can be quite slow for very large indexes.
	 */
    function build_index() {
    	$this->itemcache = 0;
        $this->items = $this->get_index_items($this->index);
        if ($this->is_multipage) {
            $this->pagecount = ceil(count($this->items) / $this->index->perpage);
            $this->pageno = ($this->pageno < 0) ? 0 : ($this->pageno >= $this->pagecount ? $this->pagecount - 1 : $this->pageno);
        }
        if ($this->has_alphalinks) {
            $this->alphalinks = $this->collect_links($this->items, $this->index->perpage);
        }
        $this->items = az_slice_array($this->items, $this->pageno, $this->pagecount, $this->index->perpage, $this->is_multipage);
	}

	/**
	 * Rebuild the index to be displayed from the cache.  This can be much faster than
	 * doing if from scratch because the index does not have to be sorted, and so only the
	 * items on the current page of the index need to be queried from the database.
	 */
	function build_index_from_cache() {
        if ($this->has_alphalinks) {
            $this->alphalinks = unserialize($this->index->linkcache);
        }
        if ($this->is_multipage) {
            $this->pagecount = ceil(count($this->itemcache) / $this->index->perpage);
            $this->pageno = ($this->pageno < 0) ? 0 : ($this->pageno >= $this->pagecount ? $this->pagecount - 1 : $this->pageno);
        }
        $ids = az_slice_array($this->itemcache, $this->pageno, $this->pagecount, $this->index->perpage, $this->is_multipage);
        $this->items = $this->get_index_items($this->index, $ids);
	}	
	
    function write_cache() {
    	global $wpdb;
        
    	$id = $this->index->id;
    	
        if (!$this->cache_disabled) {
            $itemcache = serialize($this->itemcache);
            $linkcache = serialize($this->alphalinks);
        }
    	
        $query = "UPDATE ".AZ_TABLE." SET itemcache = '$itemcache', linkcache = '$linkcache' WHERE idindex = $id";
        $rc = $wpdb->query($query);
        //az_println("fn:write_cache : query = ".$query."<br/> rc = ".$rc);
    }

    /**
     * Save the page/post id of the page the index is being shown on.  This is done
     * so that we can invalidate the page cached in WP-Super-Cache if it becomes dirty.
     *
     * @param int $idindex current index id
     * @param int $pageid page id the current index is being displayed on
     */
    function add_current_page_to_cache($idindex, $pageid) {
        //az_trace("add_current_page_to_cache: indexid = $idindex; pageid = $pageid");
        if ($pageid > 0) {
            az_acquire_cache_mutex();
            $dirtyposts = az_get_dirty_posts();
            if (!(is_array($dirtyposts['pages'][$idindex]) && in_array(intval($pageid), $dirtyposts['pages'][$idindex]))) {
                 $dirtyposts['pages'][$idindex][] = $pageid;
                 az_set_dirty_posts($dirtyposts);
                 //az_trace("   added to dirtyposts: ".implode(",", $dirtyposts['pages'][$idindex]));
            }
            az_release_cache_mutex();
        }
    }

    /**
     * Test to see if the index cache is dirty (i.e. out of date).  If there is no cache,
     * or any of the dirty items causes a change to the content and/or order of the index
     * then the index cache is dirty and needs to be updated.
     *
     * @return boolean true if the index cache is dirty and needs to be recreated
     */
    function is_dirty() {
        $dirty = false;
        $idindex = $this->index->id;
        az_acquire_cache_mutex();
    	$dirtyposts = az_get_dirty_posts($idindex);
    	if (empty($this->itemcache)) {
            // If there is no cache for the index then, by default, the index is dirty.
    	    $dirty = true;
            az_reset_dirty_posts($dirtyposts, $idindex);
    	} else if (!empty($dirtyposts)) {
    	    // If there is a cache, then check to see if the items in the dirty list invalidate it
            $dirty = $this->process_dirty_items($this->index, $this->itemcache, $dirtyposts[$idindex]);
            az_reset_dirty_posts($dirtyposts, $idindex);
    	}
        //az_dump_dirty(az_get_dirty_posts());
    	az_release_cache_mutex();
        //az_trace("is_dirty: cache: ".(!empty($this->itemcache) ? implode(",", $this->itemcache) : "empty"));
        //az_trace("is_dirty: dirty: ".(!empty($dirtyposts) ? implode(",", $dirtyposts[$idindex]) : "empty"));
        //az_trace("is_dirty: cache is ".($dirty ? "dirty" : "clean"));
        return $dirty;
	}

	/**
	 * This function tests to see if an item in the dirty list has changed in such a way 
	 * that invalidates the order and/or size of the index.  If the item is already in the
	 * index, then we put it and the previous and next items in the index (in their current
	 * order) and execute the index's sort.  If the resulting order is different, then we
	 * know that the cache is now invalid.  This function also works for the cases where
	 * an item has been deleted or removed from the index, or where a new item needs to be
	 * added. 
	 *
	 * @param $index index being tested
	 * @param $indexitems current ordered array of cached items
	 * @param $dirtyitems array of dirty items to be checked out
	 * @return boolean true if the index cache is found to be dirty
	 */
	function process_dirty_items($index, $indexitems, $dirtyitems) {
        //az_trace("process dirty items: dirty items = ".implode(",", $dirtyitems));
		$dirty = false;
		foreach ($dirtyitems as $postid) {
            $pos = array_search($postid, $indexitems);
            // Fetch the previous postid in the index (if not at the start)
            if ($pos > 0) {
                $testhit[] = $indexitems[$pos - 1];
            }
            $testhit[] = $postid;
            // Fetch the next postid in the index (if not at the end)                              
            if ($pos !== false && $pos < count($indexitems) - 1) { 
               $testhit[] = $indexitems[$pos + 1];
            }
            // Now put them through the usual retrieval and sort.
            $items = $this->get_index_items($index, $testhit, true, true);
            $result = $this->get_item_ids($items);
            // Now check to see if the sort order is the same as before or different.
            $diff = array_diff_assoc($testhit, $result);
            
            //az_trace("  test: ".implode(",", $testhit));
            //az_trace("result: ".implode(",", $result));
            //az_trace("  diff: ".implode(",", $diff));
            //az_trace(" test1: ".((count($diff) > 0 && $pos !== false) ? "true" : "false"));
            //az_trace(" test2: ".((count($diff) == 0 && $pos === false) ? "true" : "false"));
            
            // If there is a difference, the index cache is now invalid.
            if ((count($diff) > 0 && $pos !== false) || (count($diff) == 0 && $pos === false)) {
            	$dirty = true;
            	//az_trace("the result is that the cache is dirty");
              	break;
            }
            // No difference found, reset and go to the next dirty item.
            unset($testhit);
		}
        return $dirty;
	}
	
    /**
     * Get the items to be added to the index and sort them in the specified order.
     *
     * @param $index the parameters specificed for the index
     * @return items to be included in the index
     */
    function get_index_items($index, $ids = 0, $revalidate = false, $revalidatesort = false) {
        //az_trace('fn:get_index_items: pageid: '.$post->ID);

        $posts = $this->query_index_items($index, $ids, $revalidate);
        //az_println("fn:get_index_items : query = ".$query."<br/> rc = ".$posts." - count = ".count($posts));

        // Build an array of items from the query results.
        $items = array();
        $hasfilter = has_filter('azindex_heading');
        foreach ($posts as $row) {
            $item = $this->get_post_index_info($index, $row, $hasfilter);
            if ($item != null) {
                $items[] = $item;
            }
        }                     
    
        // Sort the items into alphabetical order, as specified.
        if ($items && (empty($ids) || $revalidatesort)) {
            // Both globals only used in the compare function.
            global $az_indexchars, $az_nonalphaend;
            $az_indexchars = az_is_set($index->options, 'custom-links') ? $index->customlinks : AZ_INDEXCHARS;
            
            // Determine where to put the non-alpha starting entries by setting a signed variable.
            $az_nonalphaend = az_is_set($index->options, 'non-alpha-end') ? -1 : 1;
            
            // Sort the index using the specified comparison function.
            $comparefn = (az_is_set($index->options, 'custom-sort') && trim($index->customsort) != '') ? $index->customsort : 'az_compare';
            usort($items, $comparefn);
            
            if (!$this->cache_disabled && !$revalidate) {
                // Add the ids of the sorted items to the item cache
                unset($this->itemcache);
                for ($i = 0; $i < count($items); $i++) {
                    $this->itemcache[] = $items[$i]['id'];
                }
            }
        }
        return $items;    
    }

    function get_item_ids($items) {
        $ids = array();
        if (is_array($items)) {
            for ($i = 0; $i < count($items); $i++) {
                $ids[] = $items[$i]['id'];
            }
        }
        return $ids;
    }
    
    /**
     * Query the items to be added to the index from the blog's database.
     *
     * @param $index the parameters specificed for the index
     * @return items to be included in the index
     */
    function query_index_items($index, $idlist = 0, $revalidate = false) {
        global $wpdb;
        //az_trace('fn:query_index_items: index id: '.$index->id);

        $fields = 'ID, post_title, post_excerpt, post_author';
        $query = "SELECT DISTINCT $fields FROM $wpdb->posts";
        $selection = "'post'";
        if (az_is_set($index->options, 'include-pages')) {
            if (az_is_set($index->options, 'include-pages-exclude-posts')) {
                $selection = "'page'";
            } else {
                $selection .= ",'page'";
            }
        }
        $where = "post_status = 'publish' AND post_type IN ($selection)";
        
        // We just want to validate whether the specified post(s) belong in the index.
        if ($revalidate && !empty($idlist)) {
            $ids = implode(",", $idlist);
            $where .= " AND ID IN (".$ids.")";
        }
        
        if (!$revalidate && !empty($idlist)) {
            $ids = implode(",", $idlist);
            $query .= " WHERE ID IN (".$ids.") AND $where ORDER BY FIELD(ID, $ids)";
            //az_println("query_index_items - fast query for cached items");
        } else {
            // Process the terms in the index settings.
            $terms = new az_terms($index->catids, $index->tagids, az_is_set($index->options, 'child-cats'));
            
            // Process the included and excluded categories.
            if (!empty($terms->excats)) {
                $excats .= " AND $wpdb->posts.ID NOT IN (SELECT DISTINCT object_id FROM $wpdb->term_relationships" 
                          ." INNER JOIN $wpdb->term_taxonomy AS tax3 ON ($wpdb->term_relationships.term_taxonomy_id = tax3.term_taxonomy_id"
                          ." AND tax3.taxonomy = 'category' AND tax3.term_id IN($terms->excats)))";
            }
            if (!empty($terms->incats)) {
                $query .= " INNER JOIN $wpdb->term_relationships AS rel2 ON ($wpdb->posts.ID = rel2.object_id";
                if (!empty($terms->excats)) {
                    $query .= $excats;
                }
                $query .= ") INNER JOIN $wpdb->term_taxonomy AS tax2 ON (rel2.term_taxonomy_id = tax2.term_taxonomy_id"
                         ." AND tax2.taxonomy = 'category' AND tax2.term_id IN ($terms->incats))";
            } else if (!empty($terms->excats)) {
                $query .= " INNER JOIN $wpdb->term_relationships AS rel2 ON ($wpdb->posts.ID = rel2.object_id";
                $query .= $excats.")";
            }   

            // Process the included and excluded tags.
            if (!empty($terms->extags)) {
                $extags .= " AND $wpdb->posts.ID NOT IN (SELECT DISTINCT object_id FROM $wpdb->term_relationships" 
                         ." INNER JOIN $wpdb->term_taxonomy AS tax4 ON ($wpdb->term_relationships.term_taxonomy_id = tax4.term_taxonomy_id"
                         ." AND tax4.taxonomy = 'post_tag' AND tax4.term_id IN($terms->extags)))";
            }
                
            if (!empty($terms->intags)) {
                $query .= " INNER JOIN $wpdb->term_relationships AS rel1 ON ($wpdb->posts.ID = rel1.object_id";
                if (!empty($terms->extags)) {
                    $query .= $extags;
                }
                $query .= ") INNER JOIN $wpdb->term_taxonomy AS tax1 ON (rel1.term_taxonomy_id = tax1.term_taxonomy_id"
                                              ." AND tax1.taxonomy = 'post_tag' AND tax1.term_id IN ($terms->intags))";
            } else if (!empty($terms->extags)) {
                $query .= " INNER JOIN $wpdb->term_relationships AS rel1 ON ($wpdb->posts.ID = rel1.object_id";
                $query .= $extags.")";
            }
            $query .= " WHERE $where";
        }
    
        $posts = $wpdb->get_results($query);
        //az_trace("fn:get_index_items : query = ".$query."<br/> rc = ".$posts." - count = ".count($posts));
        return $posts;    
    }
        
    function get_post_index_info($index, $row, $hasfilter = false) {
        $item = null;
        $head = trim($this->get_post_item($row->ID, $row->post_title, $row->post_excerpt, $row->post_author, $index->head, $index->headkey));
        if (!empty($head)) {
            if ($hasfilter) {
                $head = apply_filters('azindex_heading', $head);
            }
            // Remove any characters we want to ignore during the search.
            $sorthead = empty($this->ignorechars) ? $head : ltrim($head, $this->ignorechars);

            // Don't add item to the index if there is no heading.
            $item = array('id' => $row->ID,
                         'head' => $sorthead,
                         'fullhead' => $head, 
                         'subhead' => $this->get_post_item($row->ID, $row->post_title, $row->post_excerpt, $row->post_author, $index->subhead, $index->subheadkey),
                         'desc' => $this->get_post_item($row->ID, $row->post_title, $row->post_excerpt, $row->post_author, $index->desc, $index->desckey),
                        );
        }
        return $item;
    }
    
    /**
     * Get the item for a post specified in the index as the heading,
     * subheading, or description. 
     *
     * @param $type type of item to retrieve
     * @param $key item key, if the item is a custom field
     * @return item's value
     */
    function get_post_item($postid, $title, $excerpt, $author, $type, $key = false) {
        switch ($type) {
            case 'title':
                $item = $title;  
                break; 
            case 'excerpt':
                $item = $excerpt;
                break; 
            case 'author':
                $item = get_author_name($author);
                break; 
            case 'custom':
                // Only the first custom field found with key is used.
                $item = get_post_custom_values($key, $postid);
                $item = $item[0];
                break; 
        }   
        return $item;
    }

    /**
     * Collect the alphabetical links for the items in the index. Links are
     * created for the first item to begin with each letter.
     *
     * @param $items the items in the index
     * @param $perpage number of items per page
     * @return an array characters and what page of the index they are on 
    */
    function collect_links($items, $perpage) {
        for ($i = 0; $i < count($items); $i++) {
            $item = $items[$i];
            if (!empty($item)) {
                // Next entry starts with a different letter.
                if (strncasecmp($item['head'], $prevhead, 1)) {
                    $char = strtoupper($item['head'][0]);
                    $page = $this->is_multipage ? intval($i / $perpage) : 0;
                    $indexchars[] = array('char' => $char, 'page' => $page);
                }
            }
            $prevhead = $item['head'];
        }
        return $indexchars;
    }
}
/**
 * Compare function for sorting the index into alphabetical order, first using the 
 * heading, then the subheading, then the description.  Non-alpha numbers will be 
 * sorted to the top of the list (could make this an option at some point). 
 *
 * @param $in1 first index item to compare
 * @param $in2 second index item to compare
 * @return comparison result
 */
function az_compare($in1, $in2) {
    $rc = strcasecmp($in1['head'], $in2['head']);
    if ($rc == 0) {
        $rc = strcasecmp($in1['subhead'], $in2['subhead']);            
        if ($rc == 0) {
            $rc = strcasecmp($in1['desc'], $in2['desc']);            
        }
    } else {
        if (!empty($in1['head']) && !empty($in2['head'])) {
            global $az_indexchars, $az_nonalphaend;
            // Check to see if either heading starts with a non-alphanumeric character.
            $pos1 = strpos($az_indexchars, strtoupper($in1['head'][0]));
            $pos2 = strpos($az_indexchars, strtoupper($in2['head'][0]));            
            if ($pos1 === false && !($pos2 === false)) {
                $rc = -1 * $az_nonalphaend;
            } else if ($pos2 === false && !($pos1 === false)) {
                $rc = 1 * $az_nonalphaend;
            }
        }   
    }
    return $rc;
}

function az_slice_array($array, $pageno, $pagecount, $perpage, $multipage) {
        
    $start = 0;
    $length = count($array); 
    if ($multipage) {
        $length = $perpage;
        if ($pageno > 0) {
            // Keep the last item on the previous page in the front of the array if
            // we are not on the first (or only) page of the index. 
            $start = $pageno * $perpage - 1;
            $length++; 
        }
        if ($pageno < $pagecount - 1) {
            // Keep the first item on the next page at the end of the array if
            // we are not on the last (or only) page of the index. 
            $length++; 
        }
    }
    return array_slice($array, $start, $length);
}

function az_acquire_cache_mutex() {
    global $az_mutex;
    if (!isset($az_mutex)) {
        $az_mutex = new az_mutex(1, "az_sem_cache");
    }
    return $az_mutex->acquire();
}

function az_release_cache_mutex() {
    global $az_mutex;
    return $az_mutex->release();
}

class az_mutex {
    var $id;
    var $sem_id;
    var $is_acquired = false;
    var $use_flock = false;
    var $filename = '';
    var $filepointer;

    function az_mutex($id, $filename = '') {
        if (substr(PHP_OS, 0, 3) == 'WIN' || !function_exists('sem_get')) {
            $this->use_flock = true;
        }
        
        $this->id = $id;

        if ($this->use_flock) {
            if (empty($filename)){
                az_trace("Mutex:no filename specified");
            } else {
                $this->filename = $filename;
            }
        } else {
            if (!($this->sem_id = sem_get($this->id, 1))) {
                az_trace("Mutex:Error getting semaphore");
            }
        }
    }

    function acquire() {
        if ($this->use_flock) {
            if (($this->filepointer = @fopen($this->filename, "w+")) == false) {
                az_trace("Mutex:error opening mutex file<br>");
                return false;
            }
            if (flock($this->filepointer, LOCK_EX) == false) {
                az_trace("Mutex:error locking mutex file");
                return false;
            }
        } else {
            if (!sem_acquire($this->sem_id)) {
                az_trace("Mutex:error acquiring semaphore");
                return false;
            }
        }
        $this->is_acquired = true;
        return true;
    }

    function release() {
        if (!$this->is_acquired) {
            return true;
        }

        if ($this->use_flock) {
            if (flock($this->filepointer, LOCK_UN) == false) {
                az_trace("Mutex:error unlocking mutex file");
                return false;
            }
            fclose($this->filepointer);
        } else {
            if (!sem_release($this->sem_id)) {
                az_trace("Mutex:error releasing semaphore");
                return false;
            }
        }

        $this->is_acquired = false;
        return true;
    }

    function getId() {
        return $this->sem_id;
    }
}

/**
 * Class encapsulating the code for processing the terms (categories and tags) 
 * specified in the settings for an index.  Responsible for separating out 
 * the included and excluded terms and for finding all the children of the
 * specified categories.
 *
 */
class az_terms {
    var $incats; // Included category ids
    var $intags; // Included tag ids
    var $excats; // Excluded category ids
    var $extags; // Excluded tag ids
    var $tree;   // Category relationship tree extracted from database
    
    /**
     * Constructor for the az_terms class. Does all the necessary processing
     * on the terms ready to be used for querying the database for the index.
     *
     * @param string $cats comma separated string of category ids
     * @param string $tags comma separated string of tag ids
     * @param boolean $include_children true if child categories are to be included
     */
    function az_terms($cats, $tags, $include_children) {
        $terms = $this->split_terms($cats);
        $this->incats = $terms['include'];
        $this->excats = $terms['exclude'];
        $terms = $this->split_terms($tags);
        $this->intags = $terms['include'];
        $this->extags = $terms['exclude'];
        
        if ($include_children) {
            $this->tree = $this->get_cat_tree();
            $this->incats = $this->get_children($this->incats);
            $this->excats = $this->get_children($this->excats);
        }
        //az_println('incats = '.$this->incats);
        //az_println('excats = '.$this->excats);
        //az_println('intags = '.$this->intags);
        //az_println('extags = '.$this->extags);
    }
    
    /**
     * Split the list of terms into those for terms to be included 
     * and those for terms to be excluded (prefixed with ~ sign)
     *
     * @param string $termids comma separated list of term ids (categories or tags).
     * @return array containing two strings for 'include' and 'exclude' term ids. 
     */
    function split_terms($termids) {
        // Only do the split if there is a ~ sign in the string.
        if (!(strpos($termids, '~') === false)) {
            $ids = explode(',', $termids);
            foreach ($ids as $id) {
                if ($id[0] == '~') {
                    $exclude .= trim($id, '~').',';
                } else {
                    $include .= $id.',';
                }
            }
            $termlist['exclude'] = trim($exclude, ',');
            $termlist['include'] = trim($include, ',');
            
        } else if (!empty($termids)) {
            // No exclude terms, just copy into the include array. 
            $termlist['include'] = $termids;
        }
        return $termlist;
    }

    function get_children($idlist) {
        if (!empty($idlist) && !empty($this->tree)) {
            $ids = explode(",", $idlist);
            foreach ($ids as $id) {
                $children = $this->find_children(null, $id);
                if (!empty($children)) { 
                    $idlist .= ','.implode(',', $children);
                }
            }
        }
        return $idlist;
    }
    
    function find_children($children, $id) {
        foreach ($this->tree as $cat) {
            if ($cat['parent'] == $id) {
                $children[] = $cat['id'];
                $children = $this->find_children($children, $cat['id']);                                         
            }
        }
        return $children;
    }

    function get_cat_tree() {
        global $wpdb;
        $query = "SELECT * FROM $wpdb->term_taxonomy WHERE taxonomy = 'category' AND parent != 0";
        $result = $wpdb->get_results($query);
        foreach ($result as $row) {
            $tree[] = array('id' => $row->term_id, 'parent' => $row->parent);
        }
        return $tree;
    }
}

function az_dump_dirty($dirty, $pad = "") {
    foreach ($dirty as $key => $item) {
        if (is_array($item)) {
            if (is_string($key)) {
                 az_trace('   '.$key.':');
                 dump_dirty($item, "   ");   
            } else {
                 az_trace($pad."   $key = ".implode(',', $item));
            }
        } else {
            az_trace($pad."   $key = $item");
        }
    }
}
?>